from AccessControl.User import BasicUser
from OFS.SimpleItem import Item
from OFS.Folder import Folder
from Globals import HTMLFile, MessageDialog
from Products.ZPatterns.Rack import Rack, RackMountable
from string import split
import ZODB.bpthread
import types, sha

# dictionary, indexed by oid of UserSource, of dictionaries, indexed by
# user name, of dictionaries indexed by 'roles', 'domains', ...
_marker = []
_cache = {}
_cache_lock = ZODB.bpthread.allocate_lock()

class LoginUser(RackMountable, BasicUser, Item):

    """User that can login by way of a LoginManager.  Delegates actual
    behaviors to the UserSource."""

    def getUserName(self):
        return self.id

    def getRoles(self):
        return self._v_rack.rolesForUser(self)

    def getDomains(self):
        return self._v_rack.domainsForUser(self)

    def authenticate(self, password, request):
        return self._v_rack.authenticateUser(self, password, request)

    def logout(self):
        """Log out - can be called as AUTHENTICATED_USER.logout()
        Will only take effect if user has same getUserName() as
        AUTHENTICATED_USER."""
        return self._v_rack.logoutUser(self.getUserName())

    def credentialsChanged(self,credentials):
        self.cacheClear()
        self._v_rack.credentialsChanged(self,self.getUserName(),credentials)

    def cacheClear(self):
        self._v_rack.cacheClear(self.getUserName())

    # Class Metadata
    
    isALoginUser = 1

    meta_type = 'Login User'
    icon = 'misc_/LoginManager/user'




























class BasicUserSource(Rack):

    """A way of getting user objects, and implementing their innards"""

    _defaultClass = LoginUser    # make objects this class unless ZClass spec'd

    def rolesForUser(self,user):    pass
    def domainsForUser(self,user):  pass
    def authenticateUser(self,user,password,request):
            pass

    # Note: credentialsChanged and logoutUser methods are
    # required, but the default implementation is to
    # acquire these from the LoginManager where the
    # UserSource resides.  The interface for these methods
    # is defined in the LoginManager class.

    def __init__(self, id, title):
        Rack.__init__.im_func(self,id,title)
        self._role_mappings = {}

    # Cache support

    def cacheClear(self, name=None):
        global _cache, _cache_lock

        _cache_lock.acquire()
        try:
            if name:
                try:
                    del _cache[self._p_oid][name]
                except KeyError:
                    pass
            else:
                _cache[self._p_oid] = {}
        finally:
            _cache_lock.release()




    def cacheSet(self, name, key, value):
        global _cache, _cache_lock

        _cache_lock.acquire()
        try:
            _cache[self._p_oid] = usc = _cache.get(self._p_oid, {})
            usc[name] = uc = usc.get(name, {})
            uc[key] = value

        finally:
            _cache_lock.release()

    def cacheGet(self, name, key, default=None):
        global _cache #, _cache_lock

        r = default
        # _cache_lock.acquire()
        try:
            try:
                r = _cache[self._p_oid][name].get(key, default)
            except KeyError:
                pass
        finally:
            pass # _cache_lock.release()
        return r

    def cacheGetAuth(self, name, password):
        """False doesn't mean 'no', it just means we don't know 'yes'"""
        p = self.cacheGet(name, 'pass')
        if p and (sha.sha(password).digest() == p):
            return 1
        return 0

    def cacheSetAuth(self, name, password):
        self.cacheSet(name, 'pass', sha.sha(password).digest())

    # Role Mapping Support

    def _map_roles(self, roles):
        # Apply role mappings

        rd = {}

        for r in roles:
            rd[r] = 1

        for r in rd.keys():
            for mr in self._role_mappings.get(r, []):
                rd[mr] = 1

        return rd.keys()

    manage_mappingsForm = HTMLFile('www/mappings', globals())

    def role_mappings_list(self):
        """Get role mappings as from,to pairs"""
        l = []
        for k, v in self._role_mappings.items():
            for x in v:
                l.append((k, x))
        l.sort()
        return l

    def manage_mappings(self, mapping=[], REQUEST=None):
        """manage role mappings"""
        self._role_mappings = rm = {}
        for m in mapping:
            if m.mfrom and m.mto:
                if rm.has_key(m.mfrom):
                    rm[m.mfrom].append(m.mto)
                else:
                    rm[m.mfrom] = [m.mto]

        self.cacheClear()

        if REQUEST: return self.manage_main(
            self,REQUEST,update_menu=1,
            manage_tabs_message="Role mappings changed.")




    def _zclassOK(self, z):
        return hasattr(z._zclass_, 'isALoginUser')


    # Class Metadata

    manage_options_left = Rack.manage_options_left + (
            {'label':'Role Mappings',   'action':'manage_mappingsForm'},
    )

    meta_type = "Basic User Source"
    __plugin_kind__ = "User Source"
    icon = 'misc_/LoginManager/usersource'




























manage_addGenericUserSourceForm = HTMLFile('addGenericUserSource', globals())

def manage_addGenericUserSource(self, id, title='', REQUEST=None):
    """Add a Generic User Source"""
    ob = GenericUserSource(id, title)
    return self.Destination()._installPlugIn(ob,'Added Generic User Source.',REQUEST)


from LoginManager import _LoggingInUser
from AccessControl.User import nobody

def _swapAU(REQUEST,user):
    if not REQUEST.has_key('AUTHENTICATED_USER'):
        REQUEST['AUTHENTICATED_USER'] = user
        user = None
    elif user is None:
        user = REQUEST['AUTHENTICATED_USER']
        del REQUEST.other['AUTHENTICATED_USER']
    else:
        user, REQUEST['AUTHENTICATED_USER'] = REQUEST['AUTHENTICATED_USER'], user
    return user

class GenericUserSource(BasicUserSource):

    """A GUF-alike user source"""

    meta_type = "Generic User Source"

    def retrieveItem(self,name):

        exist = self.cacheGet(name, 'exist', 0)

        old_au = _swapAU(self.REQUEST,_LoggingInUser)

        if not exist and hasattr(self, 'userExists'):
            if self.userExists(self, self.REQUEST, username=name) == 1:
                exist = 1
                self.cacheSet(name, 'exist', 1)

        _swapAU(self.REQUEST, old_au)

        if exist:
            return self._RawItem(name)

        return None









    def rolesForUser(self,user):

        name = user.getUserName()

        data = self.cacheGet(name, 'roles', _marker)
        if data is not _marker:
            return data

        old_au = _swapAU(self.REQUEST,_LoggingInUser)

        try:
            data = self.userRoles(self, self.REQUEST, username=name)
            if type(data) == types.StringType:
                data = split(data)
            else:
                data = list(data)
                if len(data) > 0 and type(data[0]) != types.StringType:
                    if hasattr(data[0], 'role'):
                        data = map(lambda x: x.role, data)
                    elif hasattr(data[0], 'r'):
                        data = map(lambda x: x.r, data)
        except:
            data = ['Anonymous']
            # XXX log exception or something

        _swapAU(self.REQUEST, old_au)

        if 'Anonymous' not in data:
            data.insert(0, 'Anonymous')

        data = self._map_roles(data)

        self.cacheSet(name, 'roles', data)

        return data






    def domainsForUser(self,user):

        name = user.getUserName()

        data = self.cacheGet(name, 'domains', _marker)
        if data is not _marker:
            return data

        old_au = _swapAU(self.REQUEST,_LoggingInUser)

        try:
            data = self.userDomains(self, self.REQUEST, username=name)
            if type(data) == types.StringType:
                data = split(data)
            else:
                data = list(data)
                if len(data) > 0 and type(data[0]) != types.StringType:
                    if hasattr(data[0], 'domain'):
                        data = map(lambda x: x.domain, data)
                    elif hasattr(data[0], 'd'):
                        data = map(lambda x: x.d, data)
        except:
            data = []
            # XXX log exception or something

        _swapAU(self.REQUEST, old_au)
        self.cacheSet(name, 'domains', data)

        return data












    def authenticateUser(self,user,password,request):

        name = user.getUserName()

        if self.cacheGetAuth(name, password):
            return 1

        if not hasattr(self, 'userAuthenticate'):
            return 0

        old_au = _swapAU(self.REQUEST,_LoggingInUser)

        try:
            ok = self.userAuthenticate(
                self, request, username=name, password=password)

        finally:
            _swapAU(self.REQUEST, old_au)

        if ok:
            self.cacheSetAuth(name, password)

        return ok
